#include "config.h"
#include "buddy.h"
#include <string.h>
#include <stdint.h>

#if VMMA_DEBUG
#  include <stdio.h>
#  include <assert.h>
#endif /* VMMA_DEBUG */

#define RESERVED_BIT           0x80
#define LEVELS_MASK            0x7f
#define ADDR_TO_OFFS(_addr_)   ((((vmma_uint8_t*)(_addr_)) - buddy->memory) >> buddy->min_bits)
#define OFFS_TO_ADDR(_offs_)   (buddy->memory + (((vmma_uintptr_t)(_offs_)) << buddy->min_bits))

/* set to non-zero to check the bounds of the block in buddy_free() */
/* (by default this is set to 1) */
#define CHECK_BOUNDS_IN_FREE 1

/* set to non-zero to buddy_free() check whether the block is actually */
/* an allocated memory block or not */
/* (by default this is set to 1) */
#define CHECK_RESERVED_BIT_IN_FREE 1

struct bchunk;

typedef void*  (*pfn_alloc)(struct buddy* const buddy, size_t size);
typedef void   (*pfn_free)(struct buddy* const buddy, void* block);
typedef size_t (*pfn_msize)(struct buddy* const buddy, void* block);
typedef void   (*pfn_deinit)(struct buddy* const buddy);
typedef size_t (*pfn_get_smallest_size)(struct buddy* const buddy);
typedef size_t (*pfn_get_largest_size)(struct buddy* const buddy);
typedef size_t (*pfn_get_total_size)(struct buddy* const buddy);
typedef size_t (*pfn_get_largest_free)(struct buddy* const buddy);
typedef size_t (*pfn_get_total_free)(struct buddy* const buddy);

struct pfntab
{
	pfn_alloc             alloc;
	pfn_free              free;
	pfn_msize             msize;
	pfn_deinit            deinit;
	pfn_get_smallest_size get_smallest_size;
	pfn_get_largest_size  get_largest_size;
	pfn_get_total_size    get_total_size;
	pfn_get_largest_free  get_largest_free;
	pfn_get_total_free    get_total_free;
};

typedef const struct pfntab* C_pfntab;

static const struct pfntab pfntab_sbuddy;
static const struct pfntab pfntab_ebuddy;

struct bchunk
{
	struct bchunk* next;
	struct bchunk* prev;
};

struct buddy
{
	C_pfntab      pfntab;         /* function pointer table */
	vmma_uint8_t* memory;         /* base address of the allocatable memory */
	size_t        total_size;     /* total size of the allocatable memory (will not changed) */
	size_t        max_size;       /* maximal possible chunk size to allocate */
	size_t        min_size;       /* minimal possible chunk size to allocate */
	size_t        min_bits;       /* log2 of min_size */
	size_t        num_levels;     /* total number of levels in the heap (size of the  */
	size_t        num_minchunks;  /* total number of min-sized chunk (size of the chunks_info) */
	vmma_uint8_t* chunks_info;    /* size and reservation information about chunks */
	struct bchunk chunk_lists[1]; /* list of free chunks */
};

struct bslot
{
	vmma_uintptr_t start; /* start address from the buddy was made */
	vmma_uintptr_t end;   /* end+1 address of the used block */
	struct buddy*  buddy; /* the initialized buddy allocator */
	size_t         ublks; /* number of used blocks */
};

struct ebuddy
{
	C_pfntab       pfntab;    /* function pointer table */
	vmma_uintptr_t main;      /* memory address of the main buddy allocator */
	size_t         num_slots; /* number of used slots */
	size_t         max_slots; /* number of maximal slots can be used */
	size_t         last_slot; /* slot that is previously used in ebuddy_alloc() */
	size_t         init_size; /* size of the initial memory block */
	size_t         incr_size; /* size of the previous increment */
	unsigned int   exp_flags; /* expansion flags */
	void*          mem_ctx;   /* context to be used in 'mem_alloc' and 'mem_free' */
	void*          (*mem_alloc)(void*,size_t);
	void           (*mem_free)(void*,void*);
	struct bslot   slots[1];  /* slots, including the first one */
};

#if VMMA_DEBUG
void buddy_visualize(struct buddy* const buddy);
#endif /* VMMA_DEBUG */

/* size of the core buddy allocator state */
/* note that 'chunk_lists' has one element, so we need to subtract it */
#define sizeof_struct_buddy (sizeof(struct buddy) - sizeof(struct bchunk))

/* size of the core extendable buddy allocator state */
/* note that 'slots' has one element, so we need to subtract it */
#define sizeof_struct_ebuddy (sizeof(struct ebuddy) - sizeof(struct bslot))

static void list_init(struct bchunk* const list)
{
	list->next = list->prev = list;
}

static int list_is_empty(struct bchunk* const list)
{
	return (list==list->next);
}

static struct bchunk* list_first(struct bchunk* const list)
{
	return list->next;
}

static void list_remove(struct bchunk* const node)
{
	node->prev->next = node->next;
	node->next->prev = node->prev;
}

static void list_insert_head(struct bchunk* const list, struct bchunk* const node)
{
	node->prev = list;
	node->next = list->next;
	list->next->prev = node;
	list->next = node;
}

static void list_insert_tail(struct bchunk* const list, struct bchunk* const node)
{
	node->next = list;
	node->prev = list->prev;
	list->prev->next = node;
	list->prev = node;
}

static size_t clz64(vmma_uint64_t x)
{
	size_t n = 1;

	if(0u==x) return 64;
	if(0u==(x>>32)) { n+=32; x<<=32; }
	if(0u==(x>>48)) { n+=16; x<<=16; }
	if(0u==(x>>56)) { n+= 8; x<<= 8; }
	if(0u==(x>>60)) { n+= 4; x<<= 4; }
	if(0u==(x>>62)) { n+= 2; x<<= 2; }
	n -= (x>>63);

	return n;
}

static size_t ctz64(vmma_uint64_t x)
{
	size_t n = 1;

	if(0u==x) return 64;
	if(0u==(x & 0xFFFFFFFFul)) { n+=32; x>>=32;}
	if(0u==(x & 0x0000FFFFul)) { n+=16; x>>=16;}
	if(0u==(x & 0x000000FFul)) { n+= 8; x>>= 8;}
	if(0u==(x & 0x0000000Ful)) { n+= 4; x>>= 4;}
	if(0u==(x & 0x00000003ul)) { n+= 2; x>>= 2;}

	return n - (x & 1ul);
}

/* compute the integer log2 of a number */
/* 1024 -> 10 */
/* 1020 -> 9 */
static size_t ilog2(size_t n)
{
	if(0!=(n & (n-1))) { --n; }
	return 63 - clz64(n);
}

/* compute the bounding integer log2 of a number */
/* 1025 -> 11 */
/* 1024 -> 10 */
/* 1023 -> 10 */
static size_t ilog2_bound(size_t n)
{
	if(0==(n & (n-1))) { --n; }
	return 1 + (63 - clz64(n));
}

static struct buddy* fix_memory_alignment(void* memory, size_t* p_size)
{
	const vmma_uintptr_t alignment = sizeof(struct bchunk) - 1;
	const vmma_uintptr_t misalignment = ((vmma_uintptr_t)memory) & alignment;

	if(0!=misalignment)
	{
		/* quick sanity check */
		if((*p_size)<misalignment) { return NULL; }

		memory = ((vmma_uint8_t*)memory) + (alignment - misalignment);
		*p_size -= (alignment - misalignment);
	}

	return (struct buddy*)memory;
}

struct buddy* buddy_init(void* memory, size_t size, size_t minsize)
{
	struct buddy* const buddy = fix_memory_alignment(memory, &size);

	size_t        min_log2;
	size_t        max_log2;
	size_t        num_lists;
	size_t        remaining;
	size_t        num_min;
	vmma_uint8_t* info;

	/* more sanity check */
	if(NULL==memory || size<=sizeof_struct_buddy || 0==minsize || NULL==buddy)
	{
		return NULL;
	}

	/* minsize must be power of two */
	if(0!=(minsize & (minsize-1)))
	{
		return NULL;
	}

	/* apply minsize hard-limit */
	if(sizeof(struct bchunk)>minsize)
	{
		minsize = sizeof(struct bchunk);
	}

	/* do some magic to determine the state size */
	/* first, count the number of lists to be used */
	/* (the potentional loss here is about 1 list head) */
	min_log2  = ilog2(minsize);
	max_log2  = ilog2(size - sizeof_struct_buddy);
	num_lists = 1 + (max_log2 - min_log2);
	/* second, split the remaining space into 'minsize' chunks */
	/* and 'minsize' info (1 byte) */
	remaining = size - sizeof_struct_buddy - num_lists*sizeof(struct bchunk);
	num_min   = remaining / minsize;
	{
		/* search for the optimal num_min value */
		size_t tmp = (remaining - num_min) / minsize;
		while(num_min!=tmp && (num_min+1)!=tmp)
		{
			num_min = tmp;
			tmp     = (remaining - num_min) / minsize;
		}
	}
	/* finally, recompute the total size */
	/* (the lost space here will turn to padding later) */
	remaining = num_min * minsize;

	/* compute pointers */
	info = ((vmma_uint8_t*)buddy) + sizeof_struct_buddy + num_lists*sizeof(struct bchunk);

	buddy->pfntab        = &pfntab_sbuddy;
	buddy->memory        = (void*)((((vmma_uintptr_t)info) + ((num_min + minsize - 1))) & (~(minsize - 1)));
	buddy->total_size    = remaining;
	buddy->max_size      = ((size_t)1) << max_log2;
	buddy->min_size      = ((size_t)1) << min_log2;
	buddy->min_bits      = min_log2;
	buddy->num_levels    = num_lists;
	buddy->num_minchunks = num_min;
	buddy->chunks_info   = info;

	memset(info, 0, num_min);

	{
		size_t idx;

		for(idx=0;idx<num_lists;++idx)
		{
			list_init(buddy->chunk_lists + idx);
		}
	}

	{
		struct bchunk* const lists = buddy->chunk_lists;
		vmma_uintptr_t       addr  = (vmma_uintptr_t)buddy->memory;

		while(0<remaining)
		{
			size_t level = ctz64(addr) - min_log2; /* the level this chunk goes to */
			size_t csize = ((size_t)1) << (level + min_log2); /* chunk size */

			if(csize>remaining)
			{
				level = 64 - 1 - clz64(remaining) - min_log2;
				csize = ((size_t)1) << (level + min_log2);
			}

#if VMMA_DEBUG
			/*fprintf(stderr, "Add seg: %p -> %zu (%zu/0x%zx)\n", (void*)addr, level, csize, csize);*/
#endif /* VMMA_DEBUG */

			*info = (vmma_uint8_t)level;

			list_insert_tail(lists + level, (struct bchunk*)addr);
			addr      += csize;
			info      += ((size_t)1) << level;
			remaining -= csize;
		}
	}

#if VMMA_DEBUG
	/*buddy_visualize(buddy);*/
#endif /* VMMA_DEBUG */

	return buddy;
}

static void sbuddy_deinit(struct buddy* const buddy)
{
	/* no operation */
}

static void* sbuddy_alloc(struct buddy* const buddy, size_t size)
{
	size_t         level; /* temporary level, we do calculations in it */
	size_t         lreq;  /* level of the required size */
	vmma_uint8_t*  info;  /* pointer to the chunk info of the allocated block */
	struct bchunk* chunk; /* address of the allocated memory block */

	if(0==size)
	{
		size = buddy->min_size;
	}
	else
	if(size<buddy->min_size)
	{
		size = buddy->min_size;
	}

	level = ilog2_bound(size) - buddy->min_bits;
	lreq  = level;

	/* find the smallest block that is large enough */

	while(level<buddy->num_levels &&
		list_is_empty(buddy->chunk_lists+level))
	{
		++level;
	}

#if VMMA_DEBUG
	/*fprintf(stderr, "Alloc (%zu), lreq=%zu, level=%zu\n", size, lreq, level);*/
#endif /* VMMA_DEBUG */

	/* did we ran out of memory? */

	if(level>=buddy->num_levels ||
		list_is_empty(buddy->chunk_lists+level))
	{
		return NULL;
	}

	/* we have a block, unlink it */

	chunk = list_first(buddy->chunk_lists+level);
	info  = buddy->chunks_info + ADDR_TO_OFFS(chunk);
	list_remove(chunk);

	/* 'level' can be larger than 'lreq' */
	/* in this case, we need to split the found memory block */
	/* put all fragment blocks back to the empty lists */

	while(lreq<level)
	{
		/* size of the fragment from the oversized memory block */
		/* (size expressed in minimal-sized chunks) */
		const size_t fragment_nchunks = ((size_t)1) << (--level);

		/* locate the info of the fragment chunk */
		vmma_uint8_t* const fragment_info = info + fragment_nchunks;

		/* compute the address of the fragment chunk (half of the */
		/* current chunk, see 'level') */
		struct bchunk* const frag_chunk = (struct bchunk*)
			(((vmma_uint8_t*)chunk) + (fragment_nchunks << buddy->min_bits));

		/* set the chunk info */
		*fragment_info = level;

		/* put the free memory chunk to the list */
		list_insert_head(buddy->chunk_lists+level, frag_chunk);

#if VMMA_DEBUG
	/*fprintf(stderr, "  Frag %p, level=%zu, size=%zu (0x%zx)\n", frag_chunk, level,
		(fragment_nchunks << buddy->min_bits),
		(fragment_nchunks << buddy->min_bits));*/
#endif /* VMMA_DEBUG */
	}

	/* set info for the whole block, thus misaligned requests */
	/* can be resolved */

	memset(info, RESERVED_BIT|lreq, ((size_t)1) << lreq);

#if VMMA_DEBUG
	/*fprintf(stderr, "  Res: %p, %zu (0x%zx)\n", chunk,
		(((size_t)1) << (lreq + buddy->min_bits)),
		(((size_t)1) << (lreq + buddy->min_bits)));*/
	/*buddy_visualize(buddy);*/
#endif /* VMMA_DEBUG */

	/* we're done */

	return chunk;
}

static void sbuddy_free(struct buddy* const buddy, void* block)
{
	const vmma_uintptr_t addr = (vmma_uintptr_t)block;
	vmma_uint8_t* const  last = buddy->chunks_info + buddy->num_minchunks;

	size_t         offset;   /* offset in chunks_info */
	size_t         level;    /* current level (index list heads with) */
	vmma_uintptr_t sib_mask; /* mask to determine on which side the sibling is */
	size_t         distance; /* distance of the sibling (buddy) in bytes */
	vmma_uint8_t*  info;     /* chunk info byte of the block to be freed */

#if CHECK_BOUNDS_IN_FREE

	/* check the valid memory interval */

	if(block<(void*)buddy->memory ||
		block>=(void*)(buddy->memory + buddy->total_size))
	{
		return;
	}

#endif /* CHECK_BOUNDS_IN_FREE */

	offset = ADDR_TO_OFFS(block);

#if CHECK_RESERVED_BIT_IN_FREE

	/* is the given address represent an allocated block? */

	if(0==(RESERVED_BIT & buddy->chunks_info[offset]))
	{
		return;
	}

#endif /* CHECK_RESERVED_BIT_IN_FREE */

	/* as far as we can tell, the given address is a reserved */
	/* block, so it is possible to free it up */

	level    = buddy->chunks_info[offset] & LEVELS_MASK;
	sib_mask = buddy->min_size << level;
	distance = ((size_t)1) << level;
	info     = buddy->chunks_info + offset;

	for(;;)
	{
		vmma_uint8_t* sibling_info;

		if(0!=(addr & sib_mask))
		{
			sibling_info = info - distance;
		}
		else
		{
			sibling_info = info + distance;
		}

		/* check whether the sibling is in bounds and whether it can */
		/* be merged (has no RESERVED_BIT and both on the same level) */

		if(sibling_info>=buddy->chunks_info && /* lower bound */
			sibling_info<last && /* upper bound */
			level<buddy->num_levels && /* not on top level */
			0==(RESERVED_BIT & (*sibling_info)) && /* sibling is free */
			(LEVELS_MASK & (*sibling_info))==level) /* both on same level */
		{
			struct bchunk* const sib_node =
				(struct bchunk*)OFFS_TO_ADDR(sibling_info - buddy->chunks_info);

			if(0!=(addr & sib_mask))
			{
				/* sibling is on a lower address, we must use that lower */
				/* address in the next iteration */
				info = sibling_info;
			}

			/* remove the sibling from the empty list */
			list_remove(sib_node);

			/* "merge" current node with its sibling -- in practice: */
			/* set current level one higher, so distance to sibling */
			/* is doubled, and we must use the next bit in the sibling mask */
			++level;
			distance <<= 1;
			sib_mask <<= 1;
		}
		else
		{
			/* no further merge is possible */

			struct bchunk* const node =
				(struct bchunk*)OFFS_TO_ADDR(info - buddy->chunks_info);

			/* link into the proper list */
			list_insert_head(buddy->chunk_lists + level, node);

			/* set chunk info */
			*info = (vmma_uint8_t)level;

			/* and leave */
			return;
		}
	}
}

static size_t sbuddy_msize(struct buddy* const buddy, void* block)
{
	size_t offs;
	size_t level;

	/* first, check the valid memory interval */

	if(block<(void*)buddy->memory ||
		block>=(void*)(buddy->memory + buddy->total_size))
	{
		return 0;
	}

	/* is the given address represent an allocated block? */

	offs = ADDR_TO_OFFS(block);

	if(0==(RESERVED_BIT & buddy->chunks_info[offs]))
	{
		return 0;
	}

	/* everything matched, return the size of the memory block */

	level = buddy->chunks_info[offs] & LEVELS_MASK;

	return ((size_t)1) << (level + buddy->min_bits);
}

static size_t sbuddy_get_smallest_size(struct buddy* const buddy)
{
	return buddy->min_size;
}

static size_t sbuddy_get_largest_size(struct buddy* const buddy)
{
	return buddy->max_size;
}

static size_t sbuddy_get_total_size(struct buddy* const buddy)
{
	return buddy->total_size;
}

static size_t sbuddy_get_largest_free(struct buddy* const buddy)
{
	struct bchunk* const lists = buddy->chunk_lists;
	size_t               level = buddy->num_levels;

	while(0<(level--))
	{
		if(!list_is_empty(lists + level))
		{
			return ((size_t)1) << (level + buddy->min_bits);
		}
	}

	return 0;
}

static size_t sbuddy_get_total_free(struct buddy* const buddy)
{
	size_t sum   = 0;
	size_t level = buddy->num_levels;

	while(0<(level--))
	{
		struct bchunk* const list = buddy->chunk_lists + level;
		struct bchunk*       node = list->next;

		while(list!=node)
		{
			sum += ((size_t)1) << level;
			node = node->next;
		}
	}

	sum <<= buddy->min_bits;

	return sum;
}

static const struct pfntab pfntab_sbuddy =
{
	.alloc             = sbuddy_alloc,
	.free              = sbuddy_free,
	.msize             = sbuddy_msize,
	.deinit            = sbuddy_deinit,
	.get_smallest_size = sbuddy_get_smallest_size,
	.get_largest_size  = sbuddy_get_largest_size,
	.get_total_size    = sbuddy_get_total_size,
	.get_largest_free  = sbuddy_get_largest_free,
	.get_total_free    = sbuddy_get_total_free,
};

#define ALL_EXP_FLAGS (0 \
	| BUDDY_EXP_AUTOFREE \
	| BUDDY_EXP_ALLOCPOW2 \
	| BUDDY_EXP_PREALLOC \
	| BUDDY_EXP_INCREMENTAL \
	)

unsigned int buddy_get_supported_expansion_flags(void)
{
	return ALL_EXP_FLAGS;
}

struct buddy* buddy_init_extendable(void* memory, size_t size, size_t minsize,
	size_t max_expansions, unsigned int expansion_flags, void* ctx,
	void* (*alloc_)(void*,size_t), void (*free_)(void*,void*))
{
	const size_t         size0  = size;
	struct ebuddy* const ebuddy = (struct ebuddy*)fix_memory_alignment(memory, &size);
	struct buddy*        buddy  = NULL;

	size_t ebuddy_size = sizeof_struct_ebuddy + max_expansions*sizeof(struct bslot);
	void*  main_memory = (void*)(((vmma_uint8_t*)ebuddy) + ebuddy_size);

	if(sizeof_struct_ebuddy>size || ebuddy_size>size)
	{
		return NULL;
	}

	if(1>max_expansions)
	{
		return NULL;
	}

	size -= ebuddy_size;

	ebuddy->pfntab    = &pfntab_ebuddy;
	ebuddy->main      = (vmma_uintptr_t)main_memory;
	ebuddy->num_slots = 1;
	ebuddy->max_slots = max_expansions;
	ebuddy->last_slot = 0;
	ebuddy->init_size = size0;
	ebuddy->incr_size = size0;
	ebuddy->exp_flags = expansion_flags & ALL_EXP_FLAGS;
	ebuddy->mem_ctx   = ctx;
	ebuddy->mem_alloc = alloc_;
	ebuddy->mem_free  = free_;

	if(NULL==(buddy=buddy_init(main_memory, size, minsize)))
	{
		return NULL;
	}

	ebuddy->slots[0].start = ebuddy->main;
	ebuddy->slots[0].end   = ebuddy->main + size;
	ebuddy->slots[0].buddy = buddy;
	ebuddy->slots[0].ublks = 1; /* prevent this block from being auto-freed */

	return (struct buddy*)ebuddy;
}

static size_t compute_increment_size(const size_t minsize, const size_t incr_size, size_t size, const unsigned int flags)
{
	size_t half;

	if(minsize>size)
	{
		size = minsize;
	}

	half = size / 2;

	if(BUDDY_EXP_INCREMENTAL & flags)
	{
		if(incr_size>size)
		{
			size = incr_size;
		}
	}

	/* we have to overallocate it to have enough space to construct a buddy allocator */
	size *= 2;

	if(BUDDY_EXP_PREALLOC & flags)
	{
		if((32*1024)>half) { half = 32*1024; }
		size *= 3;
		size += half;
	}

	if(BUDDY_EXP_ALLOCPOW2 & flags)
	{
		const size_t log = (size_t)ilog2(size - 1);
		/* note: yes, "2 << log" since we subracted one in ilog2, and */
		/* looked much better than "1 << (log+1)" */
		size = ((size_t)2) << log;
	}

	return size;
}

static void* ebuddy_alloc_increase(struct ebuddy* const ebuddy, const size_t size);

static void* ebuddy_alloc(struct ebuddy* const ebuddy, size_t size)
{
	const size_t slot0 = ebuddy->last_slot;

	do
	{
		struct bslot* const slot = &ebuddy->slots[ebuddy->last_slot];

		void* const block = sbuddy_alloc(slot->buddy, size);

		if(NULL!=block)
		{
			++slot->ublks;
			return block;
		}

		if((++ebuddy->last_slot)==ebuddy->num_slots)
		{
			ebuddy->last_slot = 0;
		}
	}
	while(slot0!=ebuddy->last_slot);

	if(ebuddy->max_slots==ebuddy->num_slots)
	{
		return NULL;
	}

	return ebuddy_alloc_increase(ebuddy, size);
}

static void* ebuddy_alloc_increase(struct ebuddy* const ebuddy, const size_t size)
{
	const size_t  incr_size = compute_increment_size(ebuddy->init_size, ebuddy->incr_size, size, ebuddy->exp_flags);
	void* const   memory    = ebuddy->mem_alloc(ebuddy->mem_ctx, incr_size);
	struct buddy* buddy     = NULL;
	struct bslot* slot;
	size_t        index;
	void*         block;

	if(NULL==memory)
	{
		return NULL;
	}

	if(NULL==(buddy=buddy_init(memory, incr_size, ebuddy->slots[0].buddy->min_size)))
	{
		ebuddy->mem_free(ebuddy->mem_ctx, memory);
		return NULL;
	}

	index = ebuddy->num_slots++;

	while(0<index && ebuddy->slots[index-1].start>(vmma_uintptr_t)memory)
	{
		memcpy(&ebuddy->slots[index], &ebuddy->slots[index-1], sizeof(ebuddy->slots[0]));
		--index;
	}

	ebuddy->last_slot = index;
	ebuddy->incr_size = incr_size;
	slot = &ebuddy->slots[index];

	slot->start = (vmma_uintptr_t)memory;
	slot->end   = slot->start + incr_size;
	slot->buddy = buddy;
	slot->ublks = 0;

	if(NULL!=(block=sbuddy_alloc(slot->buddy, size)))
	{
		++slot->ublks;
		return block;
	}

	return NULL;
}

static struct bslot* find_bslot(struct ebuddy* const ebuddy, void* block)
{
	size_t lo = 0;
	size_t hi = ebuddy->num_slots;

	const vmma_uintptr_t addr = (vmma_uintptr_t)block;

	/* TODO: special case 1: only 1 slot is used */
	/* TODO: special case 2: only a few slots are used */

	while(lo<hi)
	{
		const size_t m = (lo + hi - 1) / 2;

		struct bslot* const slot = &ebuddy->slots[m];

		if(addr<slot->start)
		{
			hi = m;
		}
		else
		if(addr>=slot->end)
		{
			lo = m + 1;
		}
		else
		{
			return slot;
		}
	}

	return NULL;
}

static void ebuddy_free(struct ebuddy* const ebuddy, void* block)
{
	struct bslot* const slot = find_bslot(ebuddy, block);

	if(NULL!=slot)
	{
		sbuddy_free(slot->buddy, block);
		--slot->ublks;

		if(0==slot->ublks && (BUDDY_EXP_AUTOFREE & ebuddy->exp_flags))
		{
			void* const  memory = (void*)slot->start;
			const size_t index  = slot - &ebuddy->slots[0];

			if((index+1)!=ebuddy->num_slots)
			{
				const size_t slots_to_move = ebuddy->num_slots - index - 1;
				memmove(&ebuddy->slots[index], &ebuddy->slots[index+1], slots_to_move*sizeof(ebuddy->slots[0]));
			}

			--ebuddy->num_slots;

			if(ebuddy->last_slot>=ebuddy->num_slots)
			{
				ebuddy->last_slot = 0;
			}

			ebuddy->mem_free(ebuddy->mem_ctx, memory);
		}
	}
}

static size_t ebuddy_msize(struct ebuddy* const ebuddy, void* block)
{
	struct bslot* const slot = find_bslot(ebuddy, block);

	if(NULL!=slot)
	{
		return sbuddy_msize(slot->buddy, block);
	}

	return 0;
}

static void ebuddy_deinit(struct ebuddy* const ebuddy)
{
	size_t i;

	const vmma_uintptr_t main_addr = ebuddy->main;

	for(i=0;i<ebuddy->num_slots;++i)
	{
		if(main_addr!=ebuddy->slots[i].start)
		{
			ebuddy->mem_free(ebuddy->mem_ctx, (void*)ebuddy->slots[i].start);
		}
		else
		if(0<i)
		{
			memcpy(&ebuddy->slots[0], &ebuddy->slots[i], sizeof(ebuddy->slots[0]));
		}
	}

	ebuddy->num_slots = 1;
}

static size_t ebuddy_get_smallest_size(struct ebuddy* const ebuddy)
{
	return ebuddy->slots[0].buddy->min_size;
}

static size_t ebuddy_get_largest_size(struct ebuddy* const ebuddy)
{
	size_t res = 0;
	size_t i;

	for(i=0;i<ebuddy->num_slots;++i)
	{
		const size_t size = sbuddy_get_largest_size(ebuddy->slots[i].buddy);

		if(size>res)
		{
			res = size;
		}
	}

	return res;
}

static size_t ebuddy_get_total_size(struct ebuddy* const ebuddy)
{
	size_t res = 0;
	size_t i;

	for(i=0;i<ebuddy->num_slots;++i)
	{
		res += sbuddy_get_total_size(ebuddy->slots[i].buddy);
	}

	return res;
}

static size_t ebuddy_get_largest_free(struct ebuddy* const ebuddy)
{
	size_t res = 0;
	size_t i;

	for(i=0;i<ebuddy->num_slots;++i)
	{
		const size_t size = sbuddy_get_largest_free(ebuddy->slots[i].buddy);

		if(size>res)
		{
			res = size;
		}
	}

	return res;
}

static size_t ebuddy_get_total_free(struct ebuddy* const ebuddy)
{
	size_t res = 0;
	size_t i;

	for(i=0;i<ebuddy->num_slots;++i)
	{
		res += sbuddy_get_total_free(ebuddy->slots[i].buddy);
	}

	return res;
}

static const struct pfntab pfntab_ebuddy =
{
	.alloc             = (pfn_alloc            )ebuddy_alloc,
	.free              = (pfn_free             )ebuddy_free,
	.msize             = (pfn_msize            )ebuddy_msize,
	.deinit            = (pfn_deinit           )ebuddy_deinit,
	.get_smallest_size = (pfn_get_smallest_size)ebuddy_get_smallest_size,
	.get_largest_size  = (pfn_get_largest_size )ebuddy_get_largest_size,
	.get_total_size    = (pfn_get_total_size   )ebuddy_get_total_size,
	.get_largest_free  = (pfn_get_largest_free )ebuddy_get_largest_free,
	.get_total_free    = (pfn_get_total_free   )ebuddy_get_total_free,
};

void buddy_deinit(struct buddy* const buddy)
{
	buddy->pfntab->deinit(buddy);
}

void* buddy_alloc(struct buddy* const buddy, size_t size)
{
	return buddy->pfntab->alloc(buddy, size);
}

void buddy_free(struct buddy* const buddy, void* block)
{
	buddy->pfntab->free(buddy, block);
}

size_t buddy_msize(struct buddy* const buddy, void* block)
{
	return buddy->pfntab->msize(buddy, block);
}

size_t buddy_get_smallest_size(struct buddy* const buddy)
{
	return buddy->pfntab->get_smallest_size(buddy);
}

size_t buddy_get_largest_size(struct buddy* const buddy)
{
	return buddy->pfntab->get_largest_size(buddy);
}

size_t buddy_get_total_size(struct buddy* const buddy)
{
	return buddy->pfntab->get_total_size(buddy);
}

size_t buddy_get_largest_free(struct buddy* const buddy)
{
	return buddy->pfntab->get_largest_free(buddy);
}

size_t buddy_get_total_free(struct buddy* const buddy)
{
	return buddy->pfntab->get_total_free(buddy);
}

#if VMMA_DEBUG
void buddy_visualize(struct buddy* const buddy)
{
	vmma_uint8_t*       info = buddy->chunks_info;
	vmma_uint8_t* const end  = info + buddy->num_minchunks;
	int                 mod  = 0;

	while(info<end)
	{
		const char disp = (((*info) & RESERVED_BIT) ? '#' : ',') + mod;
		size_t     size = ((size_t)1) << ((*info) & LEVELS_MASK);

		while(0<size)
		{
			fprintf(stderr, "%c", disp);
			--size;
			++info;
		}

		mod ^= 1;
	}

	fprintf(stderr, "\n");
}
#endif /* VMMA_DEBUG */
