#include "config.h"
#include "os_dep.h"
#include <stddef.h>

#if defined(__THREOS__)
/* on Threos, use the native semaphore, because it is much faster */

#include <threos/syscall.h>

void* vmma_semget(unsigned int initial_count)
{
	const Handle sema = _CreateSemaphore(0, initial_count, 0x7fffffe, 0);

	if(INVALID_HANDLE==sema)
	{
		return NULL;
	}

	return (void*)(vmma_uintptr_t)sema;
}

#define SEMA(_sem_) ((Handle)(vmma_uintptr_t)(_sem_))

void vmma_semdel(void* sem)
{
	_DeleteObject(SEMA(sem), DELETE_ORIGINAL);
}

int vmma_semwait(void* sem, int block)
{
	intptr_t timeout;

	if(block)
	{
		timeout = SEMA_BLOCKING;
	}
	else
	{
		timeout = SEMA_NOW;
	}

	if(E_OK==_WaitSemaphore(SEMA(sem), 1, timeout))
	{
		return 0;
	}

	return -1;
}

int vmma_semsig(void* sem)
{
	if(E_OK==_SignalSemaphore(SEMA(sem), 1))
	{
		return 0;
	}

	return -1;
}

#elif VMMA_SEMA_SEMGET

/* TODO: detect this */
#include <sys/stat.h>

void* vmma_semget(unsigned int initial_count)
{
	const int semid = semget(IPC_PRIVATE, 1, IPC_CREAT|S_IRUSR|S_IWUSR);

	if(0>semid)
	{
		return NULL;
	}

	if(0>semctl(semid, 0, SETVAL, initial_count))
	{
		semctl(semid, 0, IPC_RMID);
		return NULL;
	}

	return (void*)(vmma_uintptr_t)(unsigned int)(semid+1);
}

#define SEMA(_sem_) (((int)(unsigned int)(vmma_uintptr_t)(_sem_))-1)

void vmma_semdel(void* sem)
{
	semctl(SEMA(sem), 0, IPC_RMID);
}

int vmma_semwait(void* sem, int block)
{
	const int semid = SEMA(sem);
	struct sembuf sop;

	sop.sem_num = 0;
	sop.sem_op  = -1;

	if(block)
	{
		sop.sem_flg = 0;
	}
	else
	{
		sop.sem_flg = IPC_NOWAIT;
	}

	for(;;)
	{
		if(0==semop(semid, &sop, 1))
		{
			return 0;
		}
		else
		if(EINTR!=errno)
		{
			return -1;
		}
	}

	return -1;
}

int vmma_semsig(void* sem)
{
	struct sembuf sop;

	sop.sem_num = 0;
	sop.sem_op  = 1;
	sop.sem_flg = 0;

	if(0==semop(SEMA(sem), &sop, 1))
	{
		return 0;
	}

	return -1;
}

#elif VMMA_SEMA_CREATESEMAPHORE

void* vmma_semget(unsigned int initial_count)
{
	return CreateSemaphore(NULL, initial_count, 0x7fffffe, NULL);
}

#define SEMA(_sem_) ((HANDLE)(vmma_uintptr_t)(_sem_))

void vmma_semdel(void* sem)
{
	CloseHandle(SEMA(sem));
}

int vmma_semwait(void* sem, int block)
{
	DWORD timeout;

	if(block)
	{
		timeout = INFINITE;
	}
	else
	{
		timeout = 0;
	}

	if(WAIT_OBJECT_0==WaitForSingleObject(SEMA(sem), timeout))
	{
		return 0;
	}

	return -1;
}

int vmma_semsig(void* sem)
{
	if(0==ReleaseSemaphore(SEMA(sem), 1, 0))
	{
		return 0;
	}

	return -1;
}

#else
#  error "No semaphore provider implementation found"
#endif
